@@ -41,6 +41,7 @@ string barcode_file;
 char  *adapter_1;
 char  *adapter_2;
 barcodet barcode_type    = null_null;
+bool     retain_header   = false;
 bool     filter_adapter  = false;
 bool     paired          = false;
 bool     clean           = false;
@@ -139,6 +140,8 @@ int main (int argc, char* argv[]) {
     else
 	open_files(files, barcodes, pair_1_fhs, pair_2_fhs, rem_1_fhs, rem_2_fhs, counters);
 
+    int result = 1;
+
     for (uint i = 0; i < files.size(); i++) {
 	cerr << "Processing file " << i+1 << " of " << files.size() << " [" << files[i].first.c_str() << "]\n";
 
@@ -154,26 +157,26 @@ int main (int argc, char* argv[]) {
 
 	if (paired) {
 	    if (out_file_type == FileT::gzfastq || out_file_type == FileT::gzfasta)
-	    	process_paired_reads(files[i].first, files[i].second, 
-	    			     se_bc, pe_bc,
-	    			     pair_1_gzfhs, pair_2_gzfhs, rem_1_gzfhs, rem_2_gzfhs,
-	    			     counters[files[i].first], barcode_log);
+	    	result = process_paired_reads(files[i].first, files[i].second, 
+					      se_bc, pe_bc,
+					      pair_1_gzfhs, pair_2_gzfhs, rem_1_gzfhs, rem_2_gzfhs,
+					      counters[files[i].first], barcode_log);
 	    else
-		process_paired_reads(files[i].first, files[i].second, 
-				     se_bc, pe_bc,
-				     pair_1_fhs, pair_2_fhs, rem_1_fhs, rem_2_fhs,
-				     counters[files[i].first], barcode_log);
+		result = process_paired_reads(files[i].first, files[i].second, 
+					      se_bc, pe_bc,
+					      pair_1_fhs, pair_2_fhs, rem_1_fhs, rem_2_fhs,
+					      counters[files[i].first], barcode_log);
 	} else {
 	    if (out_file_type == FileT::gzfastq || out_file_type == FileT::gzfasta)
-	    	process_reads(files[i].first, 
-	    		      se_bc, pe_bc,
-	    		      pair_1_gzfhs, 
-	    		      counters[files[i].first], barcode_log);
+	    	result = process_reads(files[i].first, 
+				       se_bc, pe_bc,
+				       pair_1_gzfhs, 
+				       counters[files[i].first], barcode_log);
 	    else
-		process_reads(files[i].first, 
-			      se_bc, pe_bc,
-			      pair_1_fhs,
-			      counters[files[i].first], barcode_log);
+		result = process_reads(files[i].first, 
+				       se_bc, pe_bc,
+				       pair_1_fhs,
+				       counters[files[i].first], barcode_log);
 	}
 	cerr <<	"  " 
 	     << counters[files[i].first]["total"] << " total reads; ";
@@ -189,6 +192,11 @@ int main (int argc, char* argv[]) {
 	    cerr << counters[files[i].first]["adapter"] << " reads with adapter sequence; ";
 	cerr << counters[files[i].first]["trimmed"]     << " trimmed reads; "
 	     << counters[files[i].first]["orphaned"]    << " orphaned paired-ends.\n";
+
+	if (!result) {
+	    cerr << "Error processing reads.\n";
+	    break;
+	}
     }
 
     cerr << "Closing files, flushing buffers...\n";
@@ -228,6 +236,8 @@ process_paired_reads(string prefix_1,
     Read  *r_1, *r_2;
     ofstream *discard_fh_1, *discard_fh_2;
 
+    int return_val = 1;
+
     string path_1 = in_path_1 + prefix_1;
     string path_2 = in_path_2 + prefix_2;
 
@@ -339,41 +349,75 @@ process_paired_reads(string prefix_1,
 	    reverse_qual(r_1->phred, r_1->inline_bc_len, overhang);
 	}
 
+	int result_1 = 1;
+	int result_2 = 1;
+
   	if (r_1->retain && r_2->retain) {
-	    (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
-		write_fastq(pair_1_fhs[bc], r_1, overhang) : 
-		write_fasta(pair_1_fhs[bc], r_1, overhang);
-	    (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ?
-		write_fastq(pair_2_fhs[bc], r_2, overhang) :
-		write_fasta(pair_2_fhs[bc], r_2, overhang);
+	    if (retain_header) {
+		result_1 = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ?
+		    write_fastq(pair_1_fhs[bc], s_1, r_1) :
+		    write_fasta(pair_1_fhs[bc], s_1, r_1);
+		result_2 = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ?
+		    write_fastq(pair_2_fhs[bc], s_2, r_2) :
+		    write_fasta(pair_2_fhs[bc], s_2, r_2);
+	    } else {
+		result_1 = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ?
+		    write_fastq(pair_1_fhs[bc], r_1, overhang) :
+		    write_fasta(pair_1_fhs[bc], r_1, overhang);
+		result_2 = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ?
+		    write_fastq(pair_2_fhs[bc], r_2, overhang) :
+		    write_fasta(pair_2_fhs[bc], r_2, overhang);
+	    }
 
 	} else if (r_1->retain && !r_2->retain) {
 	    //
 	    // Write to a remainder file.
 	    //
-	    (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
-		write_fastq(rem_1_fhs[bc], r_1, overhang) : 
-		write_fasta(rem_1_fhs[bc], r_1, overhang);
+	    if (retain_header)
+		result_1 = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
+		    write_fastq(rem_1_fhs[bc], s_1, r_1) : 
+		    write_fasta(rem_1_fhs[bc], s_1, r_1);
+	    else
+		result_1 = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
+		    write_fastq(rem_1_fhs[bc], r_1, overhang) : 
+		    write_fasta(rem_1_fhs[bc], r_1, overhang);
 
 	} else if (!r_1->retain && r_2->retain) {
 	    // Write to a remainder file.
-	    (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
-		write_fastq(rem_2_fhs[bc], r_2, overhang) : 
-		write_fasta(rem_2_fhs[bc], r_2, overhang);
+	    if (retain_header)
+		result_2 = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
+		    write_fastq(rem_2_fhs[bc], s_2, r_2) : 
+		    write_fasta(rem_2_fhs[bc], s_2, r_2);
+	    else
+		result_2 = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
+		    write_fastq(rem_2_fhs[bc], r_2, overhang) : 
+		    write_fasta(rem_2_fhs[bc], r_2, overhang);
 	}
 
+	if (!result_1 || !result_2) {
+	    cerr << "Error writing to output file for '" << bc.str() << "'\n";
+	    return_val = -1;
+	    break;
+	}
+	
 	if (discards && !r_1->retain)
-	    (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
+	    result_1 = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
 		write_fastq(discard_fh_1, s_1) : 
 		write_fasta(discard_fh_1, s_1);
 	if (discards && !r_2->retain)
-	    (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
+	    result_2 = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
 		write_fastq(discard_fh_2, s_2) : 
 		write_fasta(discard_fh_2, s_2);
 
 	delete s_1;
 	delete s_2;
 
+	if (!result_1 || !result_2) {
+	    cerr << "Error writing to discard file for '" << bc.str() << "'\n";
+	    return_val = -1;
+	    break;
+	}
+ 
 	i++;
     } while ((s_1 = fh_1->next_seq()) != NULL && 
 	     (s_2 = fh_2->next_seq()) != NULL);
@@ -387,7 +431,7 @@ process_paired_reads(string prefix_1,
     delete fh_1;
     if (interleaved == false) delete fh_2;
 
-    return 0;
+    return return_val;
 }
 
 template<typename fhType>
@@ -401,6 +445,8 @@ process_reads(string prefix,
     Read  *r;
     ofstream *discard_fh;
 
+    int return_val = 1;
+
     string path = in_path_1 + prefix;
 
     if (in_file_type == FileT::fastq)
@@ -482,18 +528,37 @@ process_reads(string prefix,
 	if (r->retain)
 	    process_singlet(r, false, barcode_log[bc], counter);
 
-	 if (r->retain)
-	     (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
-		 write_fastq(pair_1_fhs[bc], r, overhang) : 
- 		 write_fasta(pair_1_fhs[bc], r, overhang);
+	int result = 1;
 
-	 if (discards && !r->retain)
-	     (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
-		 write_fastq(discard_fh, s) : 
-		 write_fasta(discard_fh, s);
+	if (r->retain) {
+	    if (retain_header)
+		result = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
+		    write_fastq(pair_1_fhs[bc], s, r) : 
+		    write_fasta(pair_1_fhs[bc], s, r);
+	    else
+		result = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
+		    write_fastq(pair_1_fhs[bc], r, overhang) : 
+		    write_fasta(pair_1_fhs[bc], r, overhang);
+	}
 
-	 delete s;
+	if (!result) {
+	    cerr << "Error writing to output file for '" << bc.str() << "'\n";
+	    return_val = -1;
+	    break;
+	}
+	
+	if (discards && !r->retain)
+	    result = (out_file_type == FileT::fastq || out_file_type == FileT::gzfastq) ? 
+		write_fastq(discard_fh, s) : 
+		write_fasta(discard_fh, s);
 
+	if (!result) {
+	    cerr << "Error writing to discard file for '" << bc.str() << "'\n";
+	    return_val = -1;
+	    break;
+	}
+
+	delete s;
 	i++;
     } while ((s = fh->next_seq()) != NULL);
 
@@ -504,7 +569,7 @@ process_reads(string prefix,
     //
     delete fh;
 
-    return 0;
+    return return_val;
 }
 
 inline int 
@@ -778,7 +843,9 @@ int parse_command_line(int argc, char* argv[]) {
 	    {"mate-pair",            no_argument, NULL, 'M'},
 	    {"no_overhang",          no_argument, NULL, 'O'},
 	    {"filter_illumina",      no_argument, NULL, 'F'},
+	    {"retain_header",        no_argument, NULL, 'H'},
 	    {"no_read_trimming",     no_argument, NULL, 'N'},
+	    {"null_index",           no_argument, NULL, 'U'},
 	    {"index_null",           no_argument, NULL, 'u'},
 	    {"inline_null",          no_argument, NULL, 'V'},
 	    {"index_index",          no_argument, NULL, 'W'},
@@ -809,7 +876,7 @@ int parse_command_line(int argc, char* argv[]) {
 	// getopt_long stores the option index here.
 	int option_index = 0;
 
-	c = getopt_long(argc, argv, "hvcqrINFuVWxYZOPmDi:y:f:o:t:B:C:b:1:2:p:s:w:E:L:A:G:T:", long_options, &option_index);
+	c = getopt_long(argc, argv, "hHvcqrINFuVWxYZOPmDi:y:f:o:t:B:C:b:1:2:p:s:w:E:L:A:G:T:", long_options, &option_index);
 
 	// Detect the end of the options.
 	if (c == -1)
@@ -903,6 +970,9 @@ int parse_command_line(int argc, char* argv[]) {
 	case 'F':
 	    filter_illumina = true;
 	    break;
+	case 'H':
+	    retain_header = true;
+	    break;
 	case 'N':
 	    trim_reads = false;
 	    break;
@@ -914,6 +984,9 @@ int parse_command_line(int argc, char* argv[]) {
 	    if (barcode_type == null_null)
 		barcode_type = inline_null;
 	    break;
+	case 'U':
+	    barcode_type = null_index;
+	    break;
 	case 'u':
 	    barcode_type = index_null;
 	    break;
@@ -1071,17 +1144,19 @@ void help() {
 	      << "  s: set the score limit. If the average score within the sliding window drops below this value, the read is discarded (default 10).\n"
 	      << "  h: display this help messsage.\n\n"
 	      << "  Barcode options:\n"
-	      << "    --inline_null:   barcode is inline with sequence, occurs only on single-end read (default).\n"
-	      << "    --index_null:    barcode is provded in FASTQ header, occurs only on single-end read.\n"
+ 	      << "    --inline_null:   barcode is inline with sequence, occurs only on single-end read (default).\n"
+	      << "    --index_null:    barcode is provded in FASTQ header (Illumina i5 or i7 read).\n"
+	      << "    --null_index:    barcode is provded in FASTQ header (Illumina i7 read if both i5 and i7 read are provided).\n"
 	      << "    --inline_inline: barcode is inline with sequence, occurs on single and paired-end read.\n"
-	      << "    --index_index:   barcode is provded in FASTQ header, occurs on single and paired-end read.\n"
-	      << "    --inline_index:  barcode is inline with sequence on single-end read, occurs in FASTQ header for paired-end read.\n"
-	      << "    --index_inline:  barcode occurs in FASTQ header for single-end read, is inline with sequence on paired-end read.\n\n"
+	      << "    --index_index:   barcode is provded in FASTQ header (Illumina i5 and i7 reads).\n"
+	      << "    --inline_index:  barcode is inline with sequence on single-end read and occurs in FASTQ header (from either i5 or i7 read).\n"
+	      << "    --index_inline:  barcode occurs in FASTQ header (Illumina i5 or i7 read) and is inline with single-end sequence (for single-end data) on paired-end read (for paired-end data).\n\n"
 	      << "  Adapter options:\n"
 	      << "    --adapter_1 <sequence>: provide adaptor sequence that may occur on the first read for filtering.\n"
 	      << "    --adapter_2 <sequence>: provide adaptor sequence that may occur on the paired-read for filtering.\n"
 	      << "      --adapter_mm <mismatches>: number of mismatches allowed in the adapter sequence.\n\n"
 	      << "  Output options:\n"
+	      << "    --retain_header: retain unmodified FASTQ headers in the output.\n"
 	      << "    --merge: if no barcodes are specified, merge all input files into a single output file (or single pair of files).\n\n"
 	      << "  Advanced options:\n"
 	      << "    --no_read_trimming: do not trim low quality reads, just discard them.\n"
